package cz.upol.jj1.lecture11;

import java.math.BigDecimal;
import java.util.*;
import java.util.function.BiFunction;
import java.util.function.Predicate;
import java.util.stream.Stream;

public class Main {

    public static void main(String[] args) {
        teaser();

        //localClasses();
       //anonymousClasses();

        //standardInterfaces();
        //runnable();
        //userDefinedFunctionalInterfaces();

        //streamAPI();

    }

    public static void teaser(){
        List<Integer> list = new LinkedList<>();
        Random r = new Random();
        for (var i = 0; i < 5; i++){
            int num = r.nextInt(0,10);
            list.add(num);
            System.out.print(num + ", ");
        }
        System.out.println();

        int sumOfSquaresOfDistinctEvenNumbers = //hrozny nazev promenne
                list.stream()                                   //stream "kolekce"
                        .filter((var a) -> a % 2 == 0)          //jen suda
                        .distinct()                             //unikatni
                        .map( a -> a * a )                 //druhe mocniny
                        .reduce(0, (var a, var b) -> a + b);  //secist  (asi to znate jako fold)
                    //prave ty vyrazy ve forme '(arg1,...,argn) -> body' jsou anonymni funkce - lambdy

        System.out.println(sumOfSquaresOfDistinctEvenNumbers);
    }

    public static void localClasses() {
        //lokalni trida (jen "logujici" validator veku)
        class Age {
            private int age;
            Age(String age){
                try {
                    this.age = Integer.parseInt(age);
                    if (this.age < 0 || this.age > 100){
                        System.out.println("Wrong age value: " + age);
                        this.age = -1;
                        return;
                    }
                } catch (NumberFormatException e){
                    this.age = -1;
                    System.out.println("Wrong age format: " + age);
                    return;
                }
                System.out.println("Age validated: " + age);
            }

            /**
             *
             * @return age or -1 for nonvalid inputs
             */
            public int validatedAge(){
                return this.age;
            }
        }
        System.out.println(new Age("123").validatedAge());
        System.out.println(new Age("10").validatedAge());
        System.out.println(new Age("1.5").validatedAge());
        System.out.println(new Age("ADF").validatedAge());
        System.out.println(new Age("-125").validatedAge());
        System.out.println(new Age("---").validatedAge());

    }

    public static void anonymousClasses(){
        Integer[] array = new Integer[]{10,1,12,-5,4,1,2};
        // vytarime novou tridu, ktera implmentuje rozhrani Comparator
        // ani ji nedavame jmeno a rovnou vytvarime jeji (jedinou) instanci
        Arrays.sort(array, new Comparator<Integer>(){
            @Override
            public int compare(Integer i1, Integer i2) {
                return Integer.compare(i1, i2);
            }
        });
        //IDE rovnou navrhuje, abychom pouzili lambdu (to lze obecne)
        // nebo referenci na existujici metodu (to lze kvuli trivialni implementaci nasi compare)

        System.out.println(Arrays.toString(array));
        // anonymni trida odvozena od Number (abstraktni trida se 4 abstraktnimi metodami)
        numberPrinter(new Number(){
            @Override
            public int intValue() {
                return 0;
            }

            @Override
            public long longValue() {
                return 1;
            }

            @Override
            public float floatValue() {
                return 2;
            }

            @Override
            public double doubleValue() {
                return 3;
            }
        });
    }

    public static void numberPrinter(Number num){
        System.out.println(num.intValue());
        System.out.println(num.longValue());
        System.out.println(num.doubleValue());
        System.out.println(num.floatValue());
    }

    //standardni funkcni rozhrani

    public static void standardInterfaces(){
        Predicate<Integer> lambda =  // slo by i Function<Integer, Boolean> (musel by se upravit i typ metody getEqualityOnIndexFun)
                getEqualityOnIndexFun(new int[]{1,2,3},
                        new int[]{3,2,1},
                        (var x, var y) -> x == y); // casto jsou typy odvoditelne z kontextu, muzeme tedy pouzivat var

        for (var i = 0; i < 3; i++ ){
            System.out.println(lambda.test(i));
        }
    }

    public static Predicate<Integer> getEqualityOnIndexFun(int[] l1,
                                                           int[] l2,
                                                           BiFunction<Integer, Integer, Boolean> lambda){

        return (Integer index) -> lambda.apply(l1[index], l2[index]); // i tady by sel var, ale samozrejme lze psat primo typ
    }


    public static void runnable(){
        //k cemu muze byt dobry runnable?
        //napriklad k praci s vlakny (uvidime v JJ2)
        Thread t = new Thread(() -> {int i = 0;
                                     int j = 1;
                                     System.out.println(i+j);}
        );
        t.run();
    }

    public static void userDefinedFunctionalInterfaces(){
        //mame sve functional interface, odpovidajici lambdy se na nej napoji
        String result1 = callMyTernary(
                (var a, var b, var c) -> a.toString() + b.toString() + c.toString());
        String result2 = callMyTernary(
                (var a, var b, var c) -> String.format("%f", a.doubleValue() + b.doubleValue() + c.doubleValue()));
        System.out.println("Results: " + result1 + System.lineSeparator() + result2);
        MyFunctionalInterface.staticContent();
    }

    public static String callMyTernary(MyFunctionalInterface<Integer, Double, BigDecimal> fun){
        return  fun.ternary(10, 10D, BigDecimal.valueOf(10));  //volam tu definovanou signaturu - dodana lambda je navazana prave tam
    }


    //STREAM API
    public static void streamAPI(){
        //jak ziskat stream?
        Stream<Integer> values = Stream.of(1,2,3,4,5);                      //stream hodnot
        Stream<Double> arrays = Arrays.stream(new Double[]{1.0, 2.0, 3.0}); //stream z pole
        List<String> l = new ArrayList();
        l.add("Dobre");
        l.add("rano");
        l.add("Javo");
        Stream<String> collections = l.stream();                            //stream z kolekce
        Stream<Number> streams = Stream.concat(values, arrays);             //stream z vice streamu
        Stream<String> operations = collections.filter((var a) -> a.startsWith("J"));   //stream modifikaci dat z jineho streamu

        //pozor, jednou zkonzumovany stream uz nelze pouzivat
        System.out.println(operations.toList()); //je ok
        try{
            System.out.println(collections.toList()); //neni ok - uz byl zkonzumovan pri vytvareni operations
        } catch (IllegalStateException e){
            System.out.println("Tento stream uz byl zpracovan...");
            e.printStackTrace();
        }

        //koukneme do dokumentace https://docs.oracle.com/en/java/javase/17/docs/api/java.base/java/util/stream/Stream.html


        //zpatky k teaseru...
        teaser(); //ted uz by melo byt vse v nem jasne

        //agregace streamu vs iterace?
        // l: { "Dobre" "rano" "Javo"}

        for (var s : l){
            System.out.println(s.toUpperCase());
        }

        l.stream().forEach((var s) -> System.out.println(s.toUpperCase()));

        //vypada to stejne, ale chova se to uplne jinak

        //klasicky for each (for ( var s : l)) vytvari iterator (proto l musi byt Iterable) a tim prochazi strukturu - tzv externi iterace
        //stream().forEach() vyuziva znalosti vnitnri struktury - tzv interni iterace
        //     - nemusi jit jeden po druhem pomoci next(), muze byt implicitne paralelizovana
        //     - jinymi slovy zpracovavaji stream a ne puvodni kolekci a stream ma jine vlastnosti

        //navic klasicky for (var s : l) vykonava staticky kod v tele
        //kdezto stream.forEach bere lambdu


    }


}
